#include <QAction>
#include <QApplication>
#include <QGuiApplication>
#include <QJsonArray>
#include <QMenu>
#include <QMessageBox>
#include <QQmlApplicationEngine>
#include <QQmlComponent>
#include <QQmlContext>
#include <QSystemTrayIcon>
#include <QtQuick/QQuickWindow>
#include <QtWebEngine/QtWebEngine>

#include "appcore.h"
#include "desktop.h"
#include "emote.h"
#include "eventmodel.h"
#include "matrixcore.h"
#include "membermodel.h"
#include "network.h"
#include "roomlistsortmodel.h"

void AppCore::addAccount(QString profileName) {
    accounts.push_back(new MatrixCore(profileName));

    context->setContextProperty("matrix", accounts.back());

    emit accountChange();
}

void AppCore::switchAccount(QString profileName) {
    qDebug() << "switching to " << profileName;

    for (auto account : accounts) {
        if (account->profileName == profileName) {
            qDebug() << account->profileName << " = " << profileName;
            context->setContextProperty("matrix", account);
        }
    }

    emit accountChange();
}

QVariantList AppCore::getAccounts() {
    QVariantList list;
    for (auto account : accounts)
        list.push_back(QVariant::fromValue(account));

    return list;
}

int main(int argc, char* argv[]) {
    QCoreApplication::setAttribute(Qt::AA_EnableHighDpiScaling);
    QCoreApplication::setApplicationName("Trinity");
    QCoreApplication::setOrganizationName("redstrate");

    QtWebEngine::initialize();

    QApplication app2(argc, argv);

    QCommandLineParser parser;
    parser.addHelpOption();

    QCommandLineOption profileOption("profile", "Default profile to load", "profile");

    parser.addOption(profileOption);
    parser.process(app2);

    qRegisterMetaType<RequestSender>();

    // matrix
    qmlRegisterUncreatableType<AppCore>("trinity.matrix", 1, 0, "AppCore", "");
    qmlRegisterUncreatableType<EventModel>("trinity.matrix", 1, 0, "EventModel", "");
    qmlRegisterUncreatableType<MatrixCore>("trinity.matrix", 1, 0, "MatrixCore", "");
    qmlRegisterUncreatableType<Room>("trinity.matrix", 1, 0, "Room", "");
    qmlRegisterUncreatableType<Event>("trinity.matrix", 1, 0, "Event", "");
    qmlRegisterUncreatableType<Member>("trinity.matrix", 1, 0, "Member", "");
    qmlRegisterUncreatableType<Emote>("trinity.matrix", 1, 0, "Emote", "");

    // platform
    qmlRegisterUncreatableType<Desktop>("trinity.platform.desktop", 1, 0, "Desktop", "");

    QQmlApplicationEngine engine;
    QQmlContext* context = new QQmlContext(engine.rootContext(), &engine);

    AppCore* core = new AppCore();
    core->context = context;

    if (parser.isSet(profileOption)) {
        core->addAccount(parser.value(profileOption));
    } else {
        core->addAccount();
    }

    Desktop desktop;

    QSystemTrayIcon* trayIcon = new QSystemTrayIcon();

    desktop.icon = trayIcon;

    context->setContextProperty("desktop", &desktop);
    context->setContextProperty("app", core);

    QQmlComponent component(&engine);
    component.loadUrl(QUrl("qrc:/main.qml"));

    QObject* root = component.create(context);

    QAction* openAction = new QAction("Open Trinity");
    QObject::connect(openAction, &QAction::triggered, [root] {
        QQuickWindow* window = qobject_cast<QQuickWindow*>(root);
        window->setVisible(true);
    });

    QAction* hideAction = new QAction("Quit");
    QObject::connect(hideAction, &QAction::triggered, [] {
        QCoreApplication::quit();
    });

    QMenu* trayMenu = new QMenu();
    trayMenu->addAction(openAction);
    trayMenu->addAction(hideAction);

    trayIcon->setContextMenu(trayMenu);

    return app2.exec();
}
